"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing generic compressed sensing functionality.

Routine listings
----------------
indicators
    Module prodiving performance indicator determination functionality.
phase_transition
    Subpackage providing phase transition determination functionality.
reconstruction
    Subpackage providing implementations of generic reconstruction algorithms.

"""

from magni.cs import indicators
from magni.cs import phase_transition
from magni.cs import reconstruction
