"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions that return data about the system, magni, files, etc

Functions used to get various annotations and chases data are available from
this module.

See Also
--------
magni.reproducibility._annotation : Documentation of annotations
magni.reproducibility._chase : Documentation of chases

"""

from magni.reproducibility._annotation import *
from magni.reproducibility._chase import *
