"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for aiding in quest for more reproducible
research.

Routine listings
----------------
data
    Module providing functions that return various data about the system,
    magni, files, etc.
io
    Module providing input/output functions to databases containing results
    from reproducible research.

"""

from magni.reproducibility import data
from magni.reproducibility import io
