"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing input/output functionality for .mi files.

Currently, only input functionality is provided.

Routine listings
----------------
read_mi_file(path)
    Read an .mi file given by a path.

"""

from magni.afm.io._util import read_mi_file
